/*
//		Paradigm LOCATE Dynamic Memory Management Package for Borland C++ 3.0
//		Copyright (C) 1992 Paradigm Systems.  All rights reserved.
//
//		Function		
//		========
//		This module contains a complete implementation of the heap management
//		interface for the Borland C++ 3.0 compiler.  As with BC++ DOS targets,
//		the near heap is supported for the small and medium memory models with
//		a far heap supported in all memory models.
//
//		The correct definition of HEAPEND is VERY IMPORTANT since as
//		implemented in this module, Paradigm LOCATE is not aware of the heap
//		and cannot perform any overlap or other checking.  The advantage of
//		this approach is that the far heap is optimally sized to use all
//		memory from the heap base to the paragraph address defined by the
//		HEAPEND definition.
//
//		NOTE: This module requires that DOSEMU.C be part of the application.
*/

#pragma inline 							/* Uses inline assembly language */

#include	<dos.h>							/* enable(), disable(), MK_FP() */

#include	"typedefs.h"					/* General purpose type declarations */
#include	"dosemu.h"						/* Int 21h function/data prototypes */


/*
//		Define the paragrah address of the end of the far heap.  When HEAPEND
//		is 0x0000, no far heap is defined.  Note that this is the paragraph
//		address, not a memory address.  Some common settings for HEAPEND are
//		shown below:
//
//				System RAM Size		HEAPEND value
//				===============		=============
//					32K						0x0800
//					64K						0x1000
//					128K						0x2000
//					256K						0x4000
//					512K						0x8000
//
//		If you prefer a fixed size far heap, it can be accomplished by
//		declaring a character array of the appropriate size and initializing
//		the far heap variables to point to the array.  Far heaps larger than
//		64K can be accomodated by declaring the array huge.
*/

#define	HEAPEND		0x0000			/* Paragraph address of the far heap */


/*
//		The following are checks on the correctness of the heap definitions.
//		A check is made that a far heap is setup in the compact/large memory
//		models (no near heap is supported for these models).
*/

#if		(HEAPEND == 0) && (defined(__COMPACT__) || defined(__LARGE__))
#error	Must define a far heap in the compact/large memory model
#endif	/* (HEAPEND == 0) && (defined(__COMPACT__) || defined(__LARGE__)) */


#if	defined(__SMALL__) || defined(__MEDIUM__)
/*
//		Near heap management variables
*/

UINT		__brklvl ;						/* Current break level */
UINT		__heaptop ;						/* Top of the heap */
UINT		__heapbase ;					/* Base of the heap */
extern pascal UINT	edata ;			/* Address of end of static data */
#endif	/* defined(__SMALL__) || defined(__MEDIUM__)*/

#if	(HEAPEND != 0)
/*
//		Far heap management variables
*/

ULONG		_brklvl ;						/* Current break level */
ULONG		_heaptop ;						/* Top of the far heap */
ULONG		_heapbase ;						/* Base of the far heap */
UINT		_psp ;							/* PSP needed to resolve externals */

#endif	/* (HEAPEND != 0) */


static	ISRP	pOldInt21 ;				/* Original Int 21h vector */


/*
//		Function:	Int21DmmHandler	
//
//		This interrupt handler services any Int 21h calls for dynamic
//		memory management functions.  All other functions are passed on
//		to the original Int21 interrupt handler which was saved to allow
//		the use of the _chain_intr() function.
//
//		This support is only required when a far heap is implemented.
*/

#pragma	option	-O-b					/* Disable dead-code optimizations */
#pragma	option	-a-					/* Forced packed structures */
#pragma	argsused							/* Ignored the unused parameters */
static
void	interrupt	Int21DmmHandler(
				UINT	bp,	UINT	di,	UINT	si,
				UINT	ds,	UINT	es,
				UINT	dx,	UINT	cx,	UINT	bx,	UINT	ax,
				UINT	ip,	UINT	cs,	UINT	psw
			)
{
#if	(HEAPEND != 0)						/* Define when a far heap is desired */
	MCBP		ma ;							/* Pointer to the current MCB */
	MCBP		nma ;							/* Pointer to the next MCB */
	MCBP		tmp ;							/* A working pointer to an MCB */
	UINT		mcb ;							/* Application segment identifier */
#endif	/* (HEAPEND != 0) */


	/* Re-enable interrupts during processing */
	enable() ;

	/* The contents of the AH register is used to determine the service */
	switch (highbyte(ax))   {
#if	(HEAPEND != 0)						/* Define when a far heap is desired */
		case DOS_REALLOC:
			/* Make a pointer to the MCB */
			ma = (MCBP) MK_FP(es - 1, 0) ;

			/* Verify we are pointing to an MCB */
			if (ma->id != 'M' && ma->id != 'Z')   {
				/* Bad block identifier - return an error code */
				ax = 9 ;
				psw |= 1 ;
				return ;
			}

			/* Now either shrink or expand it */
			if (bx > ma->size)   {
 				/* Expand - is the next block is free/sufficiently large */
				nma = (MCBP) MK_FP(es + ma->size, 0) ;
				tmp = (MCBP) MK_FP(es + bx, 0) ;
				if (nma->owner != 0 || (ma->size + nma->size) < bx)   {
					/* Return the size block that can be allocated */
					ax = 8 ;
					bx = ma->size + ((nma->owner == 0) ? nma->size : 0) ;
					psw |= 1 ;
					return ;
				}

				/* Steal the space from the second MCB */
				mcb = bx - ma->size ;
				ma->size = bx ;

				/* Initialize the new smaller MCB */
				tmp->id = nma->id ;
				tmp->owner = nma->owner ;
				tmp->size = nma->size - mcb ;
			}
			else if (bx < ma->size)   {
				/* Release - create a new MCB holding the unallocated memory */
				tmp = (MCBP) MK_FP(es + ma->size, 0) ;
				nma = (MCBP) MK_FP(es + bx, 0) ;
				nma->size = ma->size - bx - 1 ;
				nma->owner = 0 ;
				nma->id = ma->id ;

				/* Shrink the size to the specified number */
				ma->size = bx ;

				/* Combine adjacent blocks if possible */
				if (ma->id != 'Z' && tmp->owner == 0)   {
					nma->size += tmp->size + 1 ;
					nma->id = tmp->id ;
				}

				/* Insert into the chain */
				if (ma->id == 'Z')
					ma->id = 'M' ;
			}

			/* Clear the carry to indicate success */
			psw &= 0xfffe ;
			break ;
#endif	/* (HEAPEND != 0) */

		default:
			/* All other functions are passed to the old handler for service */
			_chain_intr(pOldInt21) ;
			break ;
	}
}


/*
//		Function: InitDMM
//	
//		This initialization function is responsible for initializing the
//		near and far heaps for dynamic memory management.  Which functions
//		are installed and initialized are determined by the memory model used
//		and the value of HEAPEND.
*/

static
void	far	InitDMM(void)
{
#if	(HEAPEND != 0)						/* Define when a far heap is desired */
	MCBP		ma ;							/* Pointer to the current MCB */
	UINT		wHeapBase ;					/* Paragraph address of heap base */
	UINT		wHeapEnd ;					/* Paragraph address of heap end */
	extern	void	_terminate(int) ;	/* Used for error checking by TDREM */


	/* Save the original Int 21h handler and install ours */
	pOldInt21 = getvect(0x21) ;
	setvect(0x21, Int21DmmHandler) ;

	/* Bump the interrupt 21h handler count */
	wInt21Hanlders++ ;

	/* Set up the far heap addressing */
	asm	mov	ax, _FARHEAP
	wHeapBase = _AX ;
	ma = (MCBP) MK_FP(wHeapBase++, 0) ;

	/* Use the far heap end frame */
	wHeapEnd = HEAPEND ;

#if	defined(TDREM)
	/* Check the heap limits for obvious errors */
	if (wHeapBase > wHeapEnd)   {
		/* Call the TDREM exit routine (no return) */
		_terminate(TDE_DMMERR) ;
	}
#endif	/* defined(TDREM) */

	/* Initialize the far heap variables */
	_heapbase = _brklvl = (ULONG) MK_FP(wHeapBase, 0x0000) ;
	_heaptop = (ULONG) MK_FP(wHeapEnd, 0x0000) ;
	_psp = wHeapBase ;

	/* Initialize the far heap to be a single block */
	ma->id = 'Z' ;
	ma->owner = 0 ;
	ma->size = wHeapEnd - wHeapBase ;
#endif	/* (HEAPEND != 0) */

#if	defined(__SMALL__) || defined(__MEDIUM__)
	/* Initialize the near heap variables */
	__brklvl = __heaptop = __heapbase = (UINT) &edata ;
#endif	/* defined(__SMALL__) || defined(__MEDIUM__) */
}
#pragma	startup	InitDMM		1		/* Make an initializer */


/*
//		We need a little bit of assembly language magic to make sure
//		everything is linked in to support support the far heap.  In this case,
//		a small segment is defined to hold the first MCB in the far heap, with
//		the heap end determined by the value assigned to HEAPEND.
*/

#pragma	option	-w-
asm	_FARHEAP	segment	para	public	'FARHEAP'
asm	db		16 dup (?)
asm	_FARHEAP	ends
#pragma	option	-w.
