/*
//		Dynamic Memory Management, Version 1.00
//		Copyright (C) 1989 Paradigm Systems.  All rights reserved.
//
//		Function		
//		========
//		This module contains the code for using dynamic memory in an
//		embedded application.  The size of the heap is limited by the size
//		of DGROUP in the small and medium models and 64KB in all other
//		models.  The size of the heap used for allocation is set by the
//		HEAPSIZE macro.
//
//		The typical application of this code is to satisfy run-time library
//		routines which assume the availablility of dynamically allocated
//		memory.  Since this typically requires a small amount of dynamically
//		allocated memory, this module initializes itself on the first call
//		to malloc().
*/

#include	<stddef.h>						/* Definition of NULL */
#include	<string.h>						/* Memory manipulation functions */

#include	"typedefs.h"					/* Common typedefs and macros */

/*
//		The following structure is the memory control block (MCB) used to
//		build a list of free blocks available for allocation.  Each MCB
//		contains two fields, a pointer to the next free block and the
//		size of this free block.
*/

typedef	struct	_MCB   {
		struct _MCB	*pNext ;				/* Pointer to the next free block */
		WORD			wLen ;				/* Size of this free block */
	} MCB, *MCBP ;							/* Memory control block definition */


#define	HEAPSIZE		16384				/* Set the size of the heap (in bytes) */
#define	HEAPLEN		(HEAPSIZE / sizeof(MCB))	/* Heap size in MCBs */

static	MCBP	pFirst ;					/* Handle to the first free block */
static	MCB 	_DMMBuf[HEAPLEN] ;	/* Memory used for allocation */


/*
//		Function:	free
//
//		This function is a ROMable replacement for the run-time library
//		version of this function.
*/

void		free(
				void		*pBlock			/* Handle to the block to be released */
			)
{
	MCBP		pPrev ;						/* Previous referenced MCB */
	MCBP		pNext ;						/* Next referenced MCB */
	MCBP		pMCB ;						/* Current MCB in the free chain */

	/* Get a pointer to the block handle */
	pMCB = (MCBP) pBlock - 1 ;

	/* Check if returning a block ahead of the first free block */
	if (pFirst > pMCB)   {
		pNext = pFirst ;
		pFirst = pMCB ;
		pPrev = pMCB + pMCB->wLen ;

		/* Check if we can combine with the next block */
		if (pPrev == pNext)   {
			pMCB->wLen += pNext->wLen ;
			pMCB->pNext = pNext->pNext ;
		}
		else
			pMCB->pNext = pNext ;

		return ;
	}

	/* Scan the free list to insert the returned block */
	for (pNext = pFirst; pNext != NULL && pNext < pMCB; pPrev = pNext, pNext = pNext->pNext)   {
		if (pNext + pNext->wLen == pMCB)   {
			/* Add the newly returned block to the front of the list */
			pNext->wLen += pMCB->wLen ;
			pMCB = pNext + pNext->wLen ;

			/* Check if it can also include the next block in the list */
			if (pMCB == pNext->pNext)   {
				pNext->wLen += pMCB->wLen ;
				pNext->pNext = pMCB->pNext ;
			}
			/* All done */
			return ;
		}
	}

	/* Must add the returned block to the end of the free list */
	pPrev->pNext = pMCB ;
	pPrev = pMCB + pMCB->wLen ;

	/* Check if combined with the last block in the free list */
	if (pPrev == pNext)   {
		pMCB->wLen += pNext->wLen ;
		pMCB->pNext = pNext->pNext ;
	}
	else
		pMCB->pNext = pNext ;

	return ;
}


/*
//		Function:	malloc
//
//		This function is a ROMable replacement for the run-time library
//		version of this function.
*/

void		*malloc(
				size_t	wSize				/* Size of block to allocate */
			)
{
	MCBP		pPrev ;						/* Previous referenced MCB */
	MCBP		pMCB ;						/* Current MCB in the free chain */
	WORD		wBlocks ;					/* Size of the needed block in MCBs */

	/* Self-initialization code */
	if (pFirst == NULL)   {
		pFirst = _DMMBuf ;
		pFirst->pNext = NULL ;
		pFirst->wLen = dim(_DMMBuf) - 1 ;
	}

	/* Check for a frivolous request */
	if (wSize == 0)
		return	NULL ;

	/* Compute the size of the needed block in units of MCBs */
	wBlocks = (wSize - 1) / sizeof(MCB) + 2 ;

	/* Scan the free list for a free block meeting the size */
	for (pPrev = NULL, pMCB = pFirst; pMCB != NULL; pPrev = pMCB, pMCB = pMCB->pNext)   {
		/* Check the size of the block */
		if (pMCB->wLen >= wBlocks)   {
			/* Check if an exact fit */
			if (pMCB->wLen > wBlocks)   {
				/* Remove the allocated block from the larger free block */
				pMCB->wLen -= wBlocks ;

				/* Create a new MCB for the split block */
				pMCB += pMCB->wLen ;
				pMCB->wLen = wBlocks ;
			}
			else   {
				if (pPrev == NULL)
					/* Move the free list pointer to the next free block */
					pFirst = pMCB->pNext ;
				else
					/* Unlink the block from the free list */
					pPrev->pNext = pMCB->pNext ;
			}

			/* Return the address of the data portion of the block */
			return	(void *) (pMCB + 1) ;
		}
	}

	/* No memory available for allocation */
	return	NULL ;
}


/*
//		Function:	calloc
//
//		This function is a ROMable replacement for the run-time library
//		version of this function.
*/

void		*calloc(
				size_t	wItems,			/* Number of items */
				size_t	wSize				/* Size of block to allocate */
			)
{
	DWORD		dwSize ;						/* Size of memory to allocate */
	char		*pBuf ;						/* Pointer to the buffer */

	/* Compute the size of the buffer to allocate */
	dwSize = (DWORD) wItems * wSize ;
	pBuf = (dwSize > 0xFFFF) ? NULL : malloc((size_t) dwSize) ;

	/* Clear the newly allocated buffer */
	if (pBuf != NULL)
		memset(pBuf, 0, (WORD) dwSize) ;

	return	pBuf ;
}


/*
//		Function:	realloc
//
//		This function is a ROMable replacement for the run-time library
//		version of this function.
*/

void		*realloc(
				void		*pOldBlk,		/* Handle to the original block */
				size_t	wNewSize			/* Size of new block to allocate */
			)
{
  WORD		wOldSize ;					/* Size of the original block */
  MCBP		pOldMCB ;					/* Old MCB handle */
  void		*pNewBlk ;					/* Pointer to the new block */

	/* Extract the original block length from the MCB */
	pOldMCB = (MCBP) pOldBlk - 1 ;
	wOldSize = pOldMCB->wLen * sizeof(MCB) ;  

	/* Allocate a new block and copy the original contents */
	if ((pNewBlk = malloc(wNewSize)) != NULL)   {
		memcpy(pNewBlk, pOldBlk, wOldSize < wNewSize ? wOldSize : wNewSize) ;
		free(pOldBlk) ;
	}

	return	pNewBlk ;
}
