/*
//		Paradigm Systems Turbo Debugger Interface
//		Intel 80C186EB Port 0 Driver (polled/interrupt mode)
//
//		Copyright (C) 1990, 1991 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the internal serial port on the 
//		Intel 80C186EB microprocessor.  It can be operated in either polled
//		or serial mode (by defining the symbol COMMINT) but it must be used
//		with the custom TARGET.C since unlike other drivers, two interrupt
//		vectors are used and must not be touched by USERreset().
//
//		Function requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

#define	IOB186EB		0xff00			/* 80C186EB peripheral register base */
#include	"i80186eb.h"					/* 80C186EB definitions */


static	WORD	wS0STS ;					/* Static S0STS contents */

#if	defined(COMMINT)
#if	(COMMINT != RCV0NUM)
#error	Incorrectly specified interrupt vector - check manual
#endif	/* (COMMINT != RCV0NUM) */

static	BOOL	bTailInt = FALSE ;	/* Packet tail interrupt flag */
#endif	/* defined(COMMINT) */


/*
//		Function:	COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!((wS0STS = HLPRinportb(S0STS) | wS0STS) & SCU_RI))
		;

#if	defined(COMMINT)
	/* Set the packet tail interrupt flag */
	bTailInt = TRUE ;
#endif	/* defined(COMMINT) */

	/* Clear the flag and return the character */
	wS0STS &= ~SCU_RI ;
	return	(BYTE) HLPRinportb(S0RBUF) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept it.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!((wS0STS = HLPRinportb(S0STS) | wS0STS) & SCU_TXE))
		;

	/* Clear the xmit buffer empty flag */
	wS0STS &= ~SCU_TXE ;

	/* Write the character to the transmit buffer */
	HLPRoutportb(S0TBUF, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Check if a packet tail interrupt */
	if (bTailInt == TRUE)   {
		/* Cleanup and exit - nothing to do */
		bTailInt = FALSE ;
		HLPRoutportb(EOI, ICU_NSPEC) ;
		return ;
	}
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* defined(COMMINT) */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(S0RBUF) ;

	/* Read any pending line status */
	wS0STS = wS0STS | HLPRinportb(S0STS) ;
	wS0STS = wS0STS & 0xffbf ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the EOI for the original interrupt */
	HLPRoutportb(EOI, ICU_NSPEC) ;
#endif	/* defined(COMMINT) */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


#if	defined(COMMINT)
/*
//		Function:	COMMtxint
//
//		This function handles the transmitter interrupts when they occur by
//		issuing an EOI and simply exiting.
*/

void	interrupt	far	COMMtxint(void)
{
	HLPRoutportb(EOI, ICU_NSPEC) ;
}
#endif	/* defined(COMMINT) */


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Check for packet header character and read the packet if necessary */
	if ((wS0STS = HLPRinportb(S0STS) | wS0STS) & SCU_RI)
		COMMrcvint() ;
#endif	/* !defined(COMMINT) */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

#if	defined(COMMINT)
	/* Install the interrupt vectors */
	HLPRsetvect(RCV0NUM, (void far *) COMMrcvint) ;
	HLPRsetvect(XMT0NUM, (void far *) COMMtxint) ;
#endif	/* defined(COMMINT) */

	/* Disable receiver interrupts for the time being */
	HLPRoutportb(SCUCON, 0x0008) ;

	/* Initialize the baud rate generator and serial channel 0 */
	HLPRoutportb(B0CMP, SCU_115K) ;
	HLPRoutportb(S0CON, 0x0021) ;

	/* Flush the receive input buffer and the status */
	HLPRinportb(S0STS) ;
	HLPRinportb(S0STS) ;
	HLPRinportb(S0RBUF) ;
	HLPRinportb(S0RBUF) ;

#if	defined(COMMINT)
	/* Enable operation of the remote link */
	HLPRoutportb(SCUCON, 0) ;
#endif	/* defined(COMMINT) */
	enable() ;
}
