/*
//		Paradigm Systems Turbo Debugger Interface
//		Intel 80C186EC Serial Driver (polled/interrupt mode)
//
//		Copyright (C) 1991 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the internal serial ports on the 
//		Intel 80C186EC microprocessor.  It can be operated in either polled
//		or serial mode (by defining the symbol COMMINT) using either port 0 or
//		port 1.  Port 1 users will have some extra customization work since
//		unlike port 0, the receive interrupt can vary based on the target
//		system and the slave interrupt controller code may be unnecessary.
//
//		Function requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

#define	SCU0BASE		0x60				/* Serial channel 0 base address */
#define	SCU1BASE		0x70				/* Serial channel 1 base address */
#define	SCU0INT		5					/* Serial 0 vector offset */
#define	SCU1INT		0					/* Serial 1 vector offset (user-defined) */

#define	SCUBASE		SCU0BASE			/* SCU0BASE or SCU1BASE */
#define	IOB186EC		0xff00			/* 80C186EC peripheral register base */
#include	"i80186ec.h"					/* 80C186EC definitions */


#if	defined(COMMINT)					/* Interrupt-driven version definitions */
#if	(COMMINT == 2)						/* Test if NMI or vectored interrupts */
#define	INTBASE		0					/* Base interrupt for NMI */
#define	INTNUM		2					/* NMI interrupt vector */
#else											/* Defining a vector interrupt kernel */
#define	M_INTBASE	0x20				/* Master base interrupt vector */
#define	S_INTBASE	0x28				/* Slave base interrupt vector */
#define	INTBASE		S_INTBASE		/* Base for the serial receive interrupt */
#define	INTNUM		SCU0INT			/* Receive interrupt offset */
#endif	/* (COMMINT == 2) */

#if	(COMMINT != INTBASE + INTNUM)	/* Verify against TARGET.H definition */
#error	Incorrect interrupt vector defined in TARGET.H
#endif	/* (COMMINT != INTBASE + INTNUM) */

#endif	/* defined(COMMINT) */


static	WORD	wSxSTS ;					/* Static SxSTS contents */


/*
//		Function:	COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!((wSxSTS = HLPRinportb(SxSTS) | wSxSTS) & SCU_RI))
		;

	/* Clear the flag and return the character */
	wSxSTS &= ~SCU_RI ;
	return	(BYTE) HLPRinportb(SxRBUF) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept it.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!((wSxSTS = HLPRinportb(SxSTS) | wSxSTS) & SCU_TXE))
		;

	/* Clear the xmit buffer empty flag */
	wSxSTS &= ~SCU_TXE ;

	/* Write the character to the transmit buffer */
	HLPRoutportb(SxTBUF, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	(defined(COMMINT) && (COMMINT != 2))
	/* Disable any further serial receive interrupts */
	HLPRoutportb(S_OCW1, HLPRinportb(S_OCW1) | (1 << INTNUM)) ;
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* (defined(COMMINT) && (COMMINT != 2)) */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt and update the status */
	*p++ = wCount = HLPRinportb(SxRBUF) ;
	wSxSTS = (wSxSTS | HLPRinportb(SxSTS)) & ~SCU_RI ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)   {
			*p++ = COMMrcvchar() ;
		}

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Clear the serial interrupt request latch */
	HLPRoutport(SCUIRL, 0x0300) ;

#if	(COMMINT != 2)
	/* Issue the EOI for the original interrupt (master & slave) */
	HLPRoutportb(S_OCW2, EOI) ;
	HLPRoutportb(M_OCW2, EOI) ;

	/* Re-enable serial receive interrupts */
	HLPRoutportb(S_OCW1, HLPRinportb(S_OCW1) & ~(1 << INTNUM)) ;
#endif	/* (COMMINT != 2) */
#endif	/* defined(COMMINT) */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Check for packet header character and read the packet if necessary */
	if ((wSxSTS = HLPRinportb(SxSTS) | wSxSTS) & SCU_RI)
		COMMrcvint() ;
#endif	/* !defined(COMMINT) */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)   {
		bGotPacket = 0 ;
	}

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

#if	defined(COMMINT)
	/* Install the interrupt vectors */
	HLPRsetvect(INTBASE + INTNUM, (void far *) COMMrcvint) ;

	/* Clear any latched serial interrupts */
	HLPRoutport(SCUIRL, 0x0F00) ;

#if	(COMMINT != 2)
	/* Initialize the slave 8259A PIC */
	HLPRoutportb(S_ICW1, 0x19) ;
	HLPRoutportb(S_ICW2, S_INTBASE) ;
	HLPRoutportb(S_ICW3, 0x07) ;
	HLPRoutportb(S_ICW4, 0x01) ;
	HLPRoutportb(S_OCW1, 0xff) ;

	/* Initialize the master 8259A PIC */
	HLPRoutportb(M_ICW1, 0x19) ;
	HLPRoutportb(M_ICW2, M_INTBASE) ;
	HLPRoutportb(M_ICW3, 0x80) ;
	HLPRoutportb(M_ICW4, 0x01) ;
	HLPRoutportb(M_OCW1, 0xff) ;
#endif	/* (COMMINT != 2) */
#endif	/* defined(COMMINT) */

	/* Initialize the baud rate generator and serial channel 0 */
	HLPRoutportb(BxCMP, SCU_115K) ;
	HLPRoutportb(SxCON, 0x0021) ;

	/* Flush the receive input buffer and the status */
	HLPRinportb(SxSTS) ;
	HLPRinportb(SxSTS) ;
	HLPRinportb(SxRBUF) ;
	HLPRinportb(SxRBUF) ;

#if	defined(COMMINT) && (COMMINT != 2)
	/* Unmask the interrupt pins (master & slave) */
	HLPRoutportb(M_OCW1, HLPRinportb(M_OCW1) & 0x7f) ;
	HLPRoutportb(S_OCW1, HLPRinportb(S_OCW1) & ~(1 << INTNUM)) ;
#endif	/* defined(COMMINT) && (COMMINT != 2) */
	enable() ;
}
