/*
//		Paradigm Systems Turbo Debugger Interface
//		8250 Serial Driver (80186 interrupt controller, NMI and polled modes)
//
//		Copyright (C) 1990 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the 8250/186 devices.  By defining
//		COMMINT to be the NMI interrupt vector, an interface using the NMI
//		input for serial receive interrupts can be built.  If COMMINT is left
//		undefined, the driver defaults to polled mode.
//
//		Function requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

#define	IOB186		0xff00			/* 80186/188 peripheral register base */
#include	"i80186.h"						/* 80186 interrupt controller definitions */


/*
//		Defines required by the 8250 include file
*/
#define	SC				2					/* Word I/O scale factor */
#define	IOB8250		0x0180			/* Base address of 8250 */
#include	"i8250.h"						/* 8250 serial controller definitions */
#define	BDDIV			BAUD115K			/* Set the default baud rate */

#if	defined(COMMINT)
#if	COMMINT != 2						/* Test for NMI */
#define	INTNUM		INT1NUM			/* 80186/188 interrupt used */
#else
#define	INTNUM		2					/* NMI interrupt vector */
#endif	/* !NMI */

#if	(COMMINT != INTNUM)
#error	Incorrect COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(LINSTAT) & LS_DR))
		 ;

	/* Return it to the caller */
	return	(BYTE) HLPRinportb(RXBUF) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept it.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(LINSTAT) & LS_THRE))
		 ;

	/* Write the character to the transmit buffer */
	HLPRoutportb(TXBUF, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and the switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Disable receiver interrupts to prevent other interrupts */
	HLPRoutportb(INTENB, HLPRinportb(INTENB) & ~IE_RXINT) ;
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read any pending line status */
	HLPRinportb(LINSTAT) ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(RXBUF) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
#if	COMMINT != 2
	/* Issue the EOI to the 80186 interrupt controller */
	HLPRoutport(EOI, INTNUM) ;
#endif	/* !NMI */

	/* Re-enable receive serial interrupts */
	HLPRoutportb(INTENB, HLPRinportb(INTENB) | IE_RXINT) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Check for packet header character and read the packet if necessary */
	if (HLPRinportb(LINSTAT) & LS_DR)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect(INTNUM, (void far *) COMMrcvint) ;
#if	COMMINT != 2
	/* Initialize the 80186 interrupt controller */
	HLPRoutport(INT1CTL, 0x0018) ;
#endif	/* !NMI */
#endif	/* COMMINT */

	/* Enable access to divisor latch */
	HLPRoutportb(LINECTRL, LC_DLAB) ;
	HLPRoutportb(DIVLSB, BDDIV) ;
	HLPRoutportb(DIVMSB, 0) ;

	/* Set 8 data, 1 stop bit, no parity */
	HLPRoutportb(LINECTRL, LC_8BIT) ;

	/* Mask the serial receive interrupt output */
	HLPRoutportb(INTENB, 0) ;
	HLPRoutportb(MDMCTRL, MC_RTS | MC_DTR) ;

	/* Flush the receive input buffer */
	HLPRinportb(LINSTAT) ;
	HLPRinportb(RXBUF) ;
	
	/* Enable operation of the remote link */
#if	defined(COMMINT)
	/* Enable serial receive interrupts */
	HLPRoutportb(MDMCTRL, MC_RTS | MC_DTR | MC_INTENB) ;
	HLPRoutportb(INTENB, IE_RXINT) ;
#if	COMMINT != 2
	/* Unmask the interrupt controller input */
	HLPRoutport(INT1CTL, HLPRinport(INT1CTL) & ~0x0008) ;
#endif	/* !NMI */
#else
	HLPRoutportb(MDMCTRL, MC_RTS | MC_DTR) ;
#endif	/* COMMINT */
	enable() ;
}
