/*
//		Paradigm Systems Turbo Debugger Interface
//		8251 Serial Driver (8259A interrupt controller and polled mode)
//
//		Copyright (C) 1989, 1991 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the 8251/8259A devices.  If COMMINT
//		is left undefined, the driver defaults to polled mode.
//
//		Functions requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */


/*
//		Defines required by the i8251 include file
*/
#define	SC				1					/* Byte (1), word (2) data bus addressing */
#define	IOB8251		0xfff8			/* Base address of 8251A SIO */
#include	"i8251.h"						/* 8251 serial controller definitions */

/*
//		The following logic selects the setting of the serial receive
//		interrupt vector.
*/
#if	defined(COMMINT)
#define	IOB8259		0x0020			/* PC-compatible 8259 interrupt controller */
#include	"i8259.h"						/* 8259A interrupt controller definitions */

#define	INTBASE		8					/* Select vectors 0x08 thru 0x0f */
#define	INTNUM		0					/* Select vector offset */

#if	(COMMINT != INTBASE + INTNUM)
#error	Incorrect COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(SIOSTAT) & RXRDY))
		 ;

	/* Return it to the caller */
	return	HLPRinportb(SIODATA) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as transmitter is ready to accept.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(SIOSTAT) & TXRDY))
		 ;

	/* Write the character to the transmit buffer */
	HLPRoutportb(SIODATA, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Disable further interrupts */
	HLPRoutportb(IMR, HLPRinportb(IMR) | (1 << INTNUM)) ;
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(SIODATA) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the EOI to interrupt controller */
	HLPRoutportb(OCW2, EOI) ;

	/* Re-enable receive serial interrupts */
	HLPRoutportb(IMR, HLPRinportb(IMR) & ~(1 << INTNUM)) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Check for packet header character and read the packet if necessary */
	if (HLPRinportb(SIOSTAT) & RXRDY)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
//
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect((INTBASE + INTNUM), (void far *) COMMrcvint) ;

	/* Initialize the 8259A interrupt controller */
	HLPRoutportb(ICW1, 0x1b) ;
	HLPRoutportb(ICW2, INTBASE) ;
	HLPRoutportb(ICW4, 0x0d) ;

	/* Mask interrupt controller for serial interrupts and re-enable interrupts */
	HLPRoutportb(IMR, 0xff) ;
#endif	/* COMMINT */

	/* Add any code to initialize a baud rate generator here */
		;

	/* Force the 8251 to a known state */
	HLPRoutportb(SIOCMD, (BYTE) 0) ;
	HLPRoutportb(SIOCMD, (BYTE) 0) ;
	HLPRoutportb(SIOCMD, (BYTE) 0) ;
	
	/* Reset the 8251A */
	HLPRoutportb(SIOCMD, 0x40) ;
	
	/* Program the mode register and command registers */
	HLPRoutportb(SIOMODE, 0x4e) ;
	HLPRoutportb(SIOCMD, 0x37) ;
	
	/* Unmask the interrupt controller and enable operation of the link */
#if	defined(COMMINT)
	HLPRoutportb(IMR, HLPRinportb(IMR) & ~(1 << INTNUM)) ;
#endif	/* COMMINT */
	enable() ;
}
