/*
//		Paradigm Systems Turbo Debugger Interface
//		8256 Serial Driver (80186 interrupt controller and polled mode)
//
//		Copyright (C) 1990 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the 8256/186 devices.  If COMMINT is
//		left undefined, the driver defaults to polled mode.
//
//		Function requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

#define	IOB186		0xff00			/* 80186/188 peripheral register base */
#include	"i80186.h"						/* 80186 interrupt controller definitions */

#define	SC				2
#define	IOB8256		0x0280			/* 8256 I/O base address */
#include	"i8256.h"						/* Intel 8256 definitions */


#if	defined(COMMINT)
#define	INTBASE			0x40			/* 8256 interrupt vector base */
#define	INTNUM			4				/* Receive interrupt offset */

#if	(COMMINT != (INTBASE + INTNUM))
#error	Incorrect COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(SST) & RBRDY))
		 ;

	/* Return it to the caller */
	return	(BYTE) HLPRinportb(SRB) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as transmitter is ready to accept.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(SST) & TBRDY))
		;

	/* Write the character to the transmit buffer */
	HLPRoutportb(STB, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Disable receive interrupts */
	HLPRoutportb(RI, HLPRinportb(IR) | 0x10) ;
#else
	/* Disable interrupts in case in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	HLPRinportb(SST) ;
	*p++ = wCount = HLPRinportb(SRB) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Packet received - enable interrupts again */
	HLPRoutportb(SI, HLPRinportb(IR) | 0x10) ;

	/* Don't forget the 80188 EOI */
	HLPRoutport(EOI, INT0NUM) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	if (HLPRinportb(SST) & RBRDY)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts during initialization */
	disable() ;

#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect((INTBASE + INTNUM), (void far *) COMMrcvint) ;

	/* Initialize the 80188 hardware */
	HLPRoutport(INT0CTL, 0x0038) ;
#endif	/* COMMINT */

	/* Initialize the 8256 serial hardware */
	HLPRoutportb(CMD3, 0x81) ;
	HLPRoutportb(CMD1, 0x02) ;
	HLPRoutportb(CMD2, 0x14) ;
	HLPRoutportb(CMD3, 0x7f) ;
	HLPRoutportb(CMD3, 0xe0) ;
	HLPRoutportb(RI, 0xff) ;

	/* Enable operation of the remote link */
#if	defined(COMMINT)
	/* Unmask serial receive interrupts at the 8256 and 80188 */
	HLPRoutportb(SI, HLPRinportb(IR) | 0x10) ;
	HLPRoutport(INT0CTL, HLPRinport(INT0CTL) & ~0x0008) ;
#endif	/* COMMINT */
	enable() ;
}
