/*
//		Paradigm Systems Turbo Debugger Interface
//		8530 Serial Driver (80186 interrupt controller and polled mode)
//		Copyright (C) 1990, 1992 Paradigm Systems.  All rights reserved.
//
//		This module provides support for the 8530/186 devices.  If COMMINT
//		is left undefined, the driver defaults to polled mode.
//
//		*** NOTE ***	Check the address definitions of PCTRL and PDATA in the
//		*** NOTE ***	include file I8530.H as different systems often swap the
//		*** NOTE ***	definitions of the address inputs.
//
//		Function requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

#define	IOB186		0xff00			/* 80186/188 peripheral register base */
#include	"i80186.h"						/* 80186 interrupt controller definitions */


/*
//		Defines required by the 8530 include file
*/
#define	SC				1					/* Byte I/O scale factor */
#define	IOB8530		0x00B0			/* Base address of 8530 */
#include	"i8530.h"						/* 8530 serial controller definitions */

#if	defined(COMMINT)
#define	INTBASE		0x20				/* 8530 vector interrupt base */
#define	INTNUM		0x0c				/* Receive interrupt vector offset */

#if	(COMMINT != (INTBASE + INTNUM))
#error	Incorrect COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(PCTRL) & RXRDY))
		 ;

	/* Return it to the caller */
	return	(BYTE) HLPRinportb(PDATA) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept it.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(PCTRL) & TXRDY))
		 ;

	/* Write the character to the transmit buffer */
	HLPRoutportb(PDATA, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Disable receiver interrupts to prevent other interrupts */
	Wr8530(WR1, 0x00) ;
#else
	/* Disable any interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(PDATA) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the EOI to the 80186/188 and 8530 */
	HLPRoutport(EOI, 0x8000) ;
	HLPRoutportb(PCTRL, 0x38) ;

	/* Re-enable on receive interrupts */
	Wr8530(WR1, 0x10) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Check for packet header character and read the packet if necessary */
	if (HLPRinportb(PCTRL) & RXRDY)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

#if	defined(COMMINT)
	/* This is needed for our test hardware */
	HLPRoutport(INT0CTL, 0x0038) ;
	HLPRoutport(INT1CTL, 0x0038) ;
#endif	/* COMMINT */

	/* Issue a reset command to the 8530 channel A */
	Wr8530(WR9, 0x80) ;

	/* Initialize the 8530 serial controller */
	Wr8530(WR4, 0x44) ;
	Wr8530(WR3, 0xc0) ;
	Wr8530(WR5, 0xe2) ;
	Wr8530(WR11, 0x56) ;

	/* Baud rate generator initialization */
	Wr8530(WR12, 0x02) ;
	Wr8530(WR13, 0x00) ;
	Wr8530(WR14, 0x02) ;

	/* Enable operation of the serial channel and baud rate generator */
	Wr8530(WR3, 0xc1) ;
	Wr8530(WR5, 0xea) ;
	Wr8530(WR14, 0x03) ;

#if	defined(COMMINT)
	/* Set the interrupt vector base and initialize the vector */
	Wr8530(WR2, INTBASE) ;
	HLPRsetvect(INTBASE + INTNUM, (void far *) COMMrcvint) ;

	/* Unmask receive interrupts and set the 80186/188 to cascade mode */
	HLPRoutport(INT1CTL, HLPRinport(INT1CTL) & ~0x0008) ;
	Wr8530(WR15, 0x00) ;
	Wr8530(WR1, 0x10) ;
	Wr8530(WR9, 0x09) ;
#endif	/* COMMINT */
	enable() ;
}
