/*
//		Paradigm Systems Turbo Debugger Remote Interface (TDREM) 2.0
//		Copyright (C) 1989-1992 Paradigm Systems.  All rights reserved.
//
//		NOTE: Custom version for PC ROM BIOS applications
//
//		Function		
//		========
//		This is the module containing the target-dependent functions for
//		TDREM.  User customizable functions are
//			USERnmireset()
//			USERmaskinterrupts()
//			USERunmaskinterrupts()
//			USERtdremmask()
//			USERtdremunmask()
//			USERreset()
//			USERmembase()
//			USERmemsize()
//			USERexit()
//			UIDefaultHandler()
//
//		Also included are two functions which return the CPU and floating
//		point hardware being used in the system.
//
//			USERcputype()			Processor type determination
//			USERfputype()			Floating point hardware determination
//			USERdebugger()			Debugger determination
*/

#include "typedefs.h"					/* Common defintions and prototypes */
#include "helpers.h"						/* TDREM helper functions */
#include "target.h"						/* Target application build information */

#include	"cpu.h"							/* Pick up the processor type */
#include	"fpu.h"							/* Pick up the coprocessor type */
#include	"debug.h"						/* Pick up the debugger type */


/*
//		The follow array is used as storage for the segment descriptor table
//		that is part of the downloaded file.  Sufficient space must be allocated,
//		otherwise a load error will occur with bit 0 of AX set.
//
//		As a general rule, the size of the array can be calculated as
//	
//			sizeof(sSegD) = 64 + 12 * number of segments
//
//		In this case, the number of segments are those marked as having
//		initialized data or code.  Segments such as the stack and BSS are
//		generally not downloaded and do not need space in the segment table.
*/

char		sSegD[SEGBUF] ;					/* Segment descriptor buffer space */
WORD		wSegDSize = sizeof(sSegD) ;	/* Tell TDREM how much space exists */


/*
//		Function:  USERnmireset
//
//		This function is provided to allow the target system to reset the
//		cause of a non-maskable interrupt, if necessary.
*/

void		USERnmireset(void)
{
#if	defined(TDREMINT)
	HLPRint86(TDREMINT, UI_NMIRESET) ;
#endif	/* TDREMINT */
}


/*
//		Function:  USERmaskinterrupts
//
//		This routine is called by the TD remote package to mask off any
//		interrupts which might interfere with the operation the instruction
//		trace.
*/

void		USERmaskinterrupts(void)
{
#if	defined(TDREMINT)
	HLPRint86(TDREMINT, UI_MASK) ;
#endif	/* TDREMINT */
}


/*
//		Function:  USERunmaskinterrupts
//
//		This routine is called by the TD remote package after an instruction
//		trace to re-enable interrupts disabled by a previous call to
//		USERmaskinterrupts().
*/

void		USERunmaskinterrupts(void)
{
#if	defined(TDREMINT)
	HLPRint86(TDREMINT, UI_UNMASK) ;
#endif	/* TDREMINT */
}


/*
//		Function:  USERtdrementer
//
//		This routine is called by the TD remote package immediately after
//		gaining control from the target application, giving the system a
//		chance to mask off any interrupts which should not be serviced while
//		while in the TDREM kernel.
*/

static BOOL	bDidEnter ;					/* TRUE if application is running */

void		USERtdrementer(void)
{
#if	defined(TDREMINT)
	HLPRint86(TDREMINT, UI_ENTER) ;
#endif	/* TDREMINT */
	bDidEnter = TRUE ;
}


/*
//		Function:  USERtdremexit
//
//		This routine is called by the TDREM package just before returning
//		control to the target application, permitting the unmasking of any
//		previously masked interrupts.
//
//		This function is called before USERtdrementer() at the start of
//		the user program execution.  If data is being stored between executions
//		of USERtdrementer() and USERtdremexit(), this routine must take this
//		situation into account.
*/

void		USERtdremexit(void)
{
	/* Check if whether the vector has been initialized */
	if (bDidEnter == TRUE)   {
#if	defined(TDREMINT)
		HLPRint86(TDREMINT, UI_EXIT) ;
#endif	/* TDREMINT */
	}
}


#if	defined(TDREMINT)
/*
//		Function:  UIDefaultHandler
//
//		This is the default interrupt handler for the user interface.  An
//		application desiring to modify the operation of the TDREM kernel
//		must replace this handler with one of its own choosing.
*/

void	interrupt far	UIDefaultHandler(void)
{
	return ;
}
#endif	/* TDREMINT */


/*
//		Function:  USERreset
//
//		This routine is called by the TD remote package when communication is
//		first made or when a program is loaded.  This function gives the user
//		the chance to make sure that the system I/O state closely matches a
//		system that has just been reset or to disable any interrupts which
//		might cause the target system to hang when an interrupt vector is
//		overwritten.
*/

void		USERreset(void)
{
	WORD		i ;							/* Loop counter */
	WORD		wCommInt = 256 ;			/* TDREM comm interrupt vector number */
#if	defined(TDREMINT)
	static BOOL	bInited = FALSE ;		/* TRUE if executed at least once */
#endif	/* TDREMINT */
	
/*
//		*** NOTE ***
//		Disable any active interrupts before re-installing the default
//		set of interrupt handlers.  This action is application and target
//		system dependent and failure to do so can cause the kernel to crash
//		when an active interrupt vector is over-written.
//		*** NOTE ***
*/

#if	defined(TDREMINT)
	/* Call the user handler after the first pass thru the code */
	if (bInited == TRUE)
		HLPRint86(TDREMINT, UI_RESET) ;
#endif	/* TDREMINT */

	/* Set up the default interrupt handlers while avoiding the BIOS */
	for (i = 0; i < 256; i++)   {
		/* Check if the vector contains the TDREM interrupt handler */
		if (HLPRgetvect(i) == COMMrcvint)   {
			wCommInt = i ;
			bIntDriven = TRUE ;
		}
		else if (i >= 32)
			HLPRsetvect(i, (void far *) Prognovector) ;
	}

	/* Install the interrupt handlers for NMI and exceptions */
	HLPRsetvect(0x00, (void far *) Progzerodivide) ;
	if (wCommInt != 2)
		HLPRsetvect(0x02, (void far *) Prognmientry) ;

#if	defined(TDREMINT)
	/* Install the default kernel interface interrupt handler */
	HLPRsetvect(TDREMINT, (void far *) UIDefaultHandler) ;
	bInited = TRUE ;
#endif	/* TDREMINT */

	/* Mark the application as not being loaded */
	bDidEnter = FALSE ;
}


/*
//		Function:  USERmembase
//
//		This function is used to set the default segment addresses when
//		no program is loaded.
*/

WORD		USERmembase(
				WORD		wDefSeg			/* Default user segment */
			)
{
	/* Return something different if the default is not desired */
	return	wDefSeg ;
}


/*
//		Function:  USERmemsize
//
//		This routine is called to compute the amount of available memory in
//		the system.  The result can be viewed in the Get Info command under
//		the Files menu.
*/

WORD		USERmemsize(void)
{
	/* Return the size of memory in KB */
	return	0 ;
}


/*
//		Function:  USERexit
//
//		This function is called by the kernel when TD is shutting down the
//		target system.
//
//		If it is desirable to return to some other program, the TDREMA.ASM
//		startup code and this function should be modified to save the return
//		address and exit to the caller.
*/

void		USERexit(void)
{
#if	defined(TDREMINT)
		HLPRint86(TDREMINT, UI_BYE) ;
#endif	/* TDREMINT */
}


/*
//		Function:  USERcputype
//
//		Return the processor type which can be found in the CPU.H header
//		file.  A list of the valid processor hardware IDs can be
//		found in the TARGET.H header file.
*/

BYTE		USERcputype(void)
{
	return	CPU_TYPE ;
}


/*
//		Function:	USERfputype
//
//		Return the current coprocessor type which can be found in the FPU.H
//		header file.  A list of the valid floating point hardware IDs can be
//		found in the TARGET.H header file.
*/

BYTE		USERfputype(void)
{
	return	FPU_TYPE ;
}


/*
//		Function:	USERdebugger
//
//		Return the type of debugger being used so the correct protocol can be
//		supported.
*/

BYTE		USERdebugger(void)
{
	return	DBG_TYPE ;
}
