/*
//		Paradigm Systems Turbo Debugger Remote Interface
//		V25+/V35+ Serial Driver (V25+/V35+ interrupt controller and polled mode)
//
//		Copyright (C) 1991 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for interfacing the TD remote package
//		to the NEC V25+/V35+ serial port.  Note that this module requires access
//		to the special function registers and caches a pointer.  This module
//		includes some extra code which reloads the pointer under the assumption
//		it may have been changed by the application.
//
//		A polled version of this driver can be built by not defining the
//		symbol COMMINT in the TARGET.H header file.
//
//		Functions requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */
#include	"v25pv35p.h"					/* V25+/V35+ hardware definitions */


#define	SIOPORT			0				/* 0 or 1 for V25+/V35+ serial ports */
#if	(SIOPORT == 0)
#define	SIOBASE			0x60			/* Channel 0 offset with in the SFRs */
#define	SIORCVINT		SIO0RCV		/* Channel 0 receive interrupt vector number */
#else
#define	SIOBASE			0x70			/* Channel 1 offset with in the SFRs */
#define	SIORCVINT		SIO1RCV		/* Channel 1 receive interrupt vector number */
#endif


#if	defined(COMMINT)
#if	(COMMINT != SIORCVINT)			/* Checking the interrupt vector */
#error	Uninitialized COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Baud rate divisors are based on the clock speed of the V25+/V35+
//		target hardware and must be adjusted for each system.  These
//		compile time calculations take care of this chore automatically,
//		assuming that the BDSCC prescalar is correctly set.
*/
#define	BAUDRATE		115200UL			/* Desired baud rate */
#define	XTAL_OSC		16000000UL		/* Crystal OSC frequency */
#define	BAUDCLK		(XTAL_OSC/4)	/* Baud Clock */

#define	BDSCC			0
#define	BDBRG0		(BAUDCLK / (BAUDRATE * (1 << BDSCC)))
#define	BDBRG1		(BDBRG0 + 1)
#define	MBRG0			(BDBRG0 * BAUDRATE)
#define	MBRG1			(BDBRG1 * BAUDRATE)
#if ((BAUDCLK - MBRG0) < (MBRG1 - BAUDCLK))
#define	BDBRG			(BAUDCLK / (BAUDRATE * (1 << BDSCC)))
#else
#define	BDBRG			((BAUDCLK / (BAUDRATE * (1 << BDSCC))) + 1)
#endif

/* Test that the result is reasonable and within range */
#if	!defined(MSC51)
#if	(BDBRG < 2) || (BDBRG > 255) || (BDSCC > 8)
#error	Illegal baud rate initialization values - try adjusting BDSCC
#endif	/* Baud rate check */
#endif	/* !MSC51 */


static volatile V25SIOP	pSIO ;		/* Handle to V25+/V35+ serial I/O */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Make a pointer to the serial SFRs */
	pSIO = (V25SIOP) MK_FP((*(BYTE_PTR) MK_FP(0xffff, 0xf)) << 8, 0x0f00 + SIOBASE) ;

	/* Wait for a character to appear */
	while (!(pSIO->scs & 0x10))
		;

#if	defined(COMMINT)
	/* Clear the interrupt request flag */
	pSIO->sric &= ~INTFLAG ;
#endif	/* COMMINT */

	/* Return it to the caller */
	return	pSIO->rxb ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as transmitter is ready to accept.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Make a pointer to the serial SFRs */
	pSIO = (V25SIOP) MK_FP((*(BYTE_PTR) MK_FP(0xffff, 0xf)) << 8, 0x0f00 + SIOBASE) ;

	/* Wait for the transmit buffer empty flag to set */
	while (!(pSIO->scs & 0x20))
		;

	/* Output the character to the transmitter output buffer */
	pSIO->txb = (BYTE) bTXchar ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	!defined(COMMINT)
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Make a pointer to the serial SFRs */
	pSIO = (V25SIOP) MK_FP((*(BYTE_PTR) MK_FP(0xffff, 0xf)) << 8, 0x0f00 + SIOBASE) ;

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read character that caused the interrupt */
	*p++ = wCount = (WORD) pSIO->rxb ;

	/* Clear the receive character interrupt flag */
	pSIO->sric &= ~INTFLAG ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the FINT to release the interrupt controller */
	fint() ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;

#if	!defined(COMMINT)
	/* Make a pointer to the serial SFRs */
	pSIO = (V25SIOP) MK_FP((*(BYTE_PTR) MK_FP(0xffff, 0xf)) << 8, 0x0f00 + SIOBASE) ;

	/* Test for a received character */
	if (pSIO->scs & 0x10)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

	/* Point to the IDB register and make a pointer to the SFRs */
	pSIO = (V25SIOP) MK_FP((*(BYTE_PTR) MK_FP(0xffff, 0xf)) << 8, 0x0f00 + SIOBASE) ;

#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect(SIORCVINT, (void far *) COMMrcvint) ;
#endif	/* COMMINT */

	/* Setup the mode and baud rate generator registers */
	pSIO->scm = 0xc9 ;
	pSIO->brg = (BYTE) BDBRG ;
	pSIO->scc = (BYTE) BDSCC ;

	/* Enable operation of the remote link */
#if	defined(COMMINT)
	/* Enable receive interrupts */
	pSIO->sric &= ~INTMASK ;
#endif	/* COMMINT */
	enable() ;
}
