/*
//		Paradigm Systems Turbo Debugger Interface
//		V40/V50 Serial Driver (V40/V50 ICU and polled mode)
//
//		Copyright (C) 1989, 1990 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the NEC V40/V50 SCU (serial control
//		unit) and ICU (interrupt control unit) or in polled mode if the symbol
//		COMMINT is not defined.
//
//		Functions requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

/*
//		Internal peripheral addresses which must be defined for the V40/V50
//		include file.
*/
#define	OPHA_BASE	0x00				/* OPHA initial value */
#define	IULA_BASE	0x20				/* Offset of V40/V50 ICU */
#define	SULA_BASE	0x60				/* Offset of V40/V50 SCU */
#define	TULA_BASE	0x40				/* Offset of V40/V50 TCU */
#define	SC				1					/* I/O port address scale (V40=1, V50=2) */
#include	"v40v50.h"						/* NEC V40/V50 definitions */

/*
//		Define the default peripheral register values.
*/
#define	OPCN_INIT	0x06				/* SCU enabled using INT1 */
#define	OPSEL_INIT	0x0c				/* SCU and TCU enabled */
#define	WMB_INIT		0x11				/* 64KB RAM, 64KB EPROM */
#define	WCY1_INIT	0x80				/* Memory and I/O wait states */
#define	WCY2_INIT	0x00				/* Refresh/DMA wait states */
#define	TCKS_INIT	0x00				/* All timers use the internal clock */

#define	ICU_IIW1		0x1b				/* Level-sensitive interrupts */
#define	ICU_IIW4		0x0d				/* Normal nesting mode */

#define	T1_MODE		0x76				/* Mode 3, two bytes, binary */
#define	BAUD9600		26					/*  9.6KB @ 8Mhz */

#define	SCU_MODE		0x4e				/* 8 data, 1 stop, no parity, 16x clock */
#define	SCU_CMD		0x37				/* Enable xmit/rcv */
#define	BRDIV			BAUD9600			/* Selected baud rate */


#if	defined(COMMINT)
#define	INTBASE		8					/* Hardware interrupt for bus interrupt 0 */
#define	INTNUM		1					/* V40/V50 serial interrupt */

#if	(COMMINT != INTBASE + INTNUM)
#error	Uninitialized COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(SST) & RBRDY))
		 ;

	/* Return it to the caller */
	return	HLPRinportb(SRB) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(SST) & TBRDY))
		;

	/* Write the character to the transmit buffer */
	HLPRoutportb(STB, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Mask serial receive interrupts */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) | 0x01) ;
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(SRB) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the EOI to interrupt controller */
	HLPRoutportb(IPFW, EOI) ;

	/* Unmask serial receive interrupts */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) & 0xfe) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;


#if	!defined(COMMINT)
	/* Check if a character is available */
	if (HLPRinportb(SST) & RBRDY)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

	/* Initialize the V40/V50 system I/O registers */
	HLPRoutportb(OPSEL, OPSEL_INIT) ;
	HLPRoutportb(OPCN, OPCN_INIT) ;
	HLPRoutportb(TCKS, TCKS_INIT) ;

	/* Set the I/O port addresses of internal peripherals */
	HLPRoutportb(OPHA, OPHA_BASE) ;
	HLPRoutportb(SULA, SULA_BASE) ;
	HLPRoutportb(TULA, TULA_BASE) ;
#if	defined(COMMINT)
	HLPRoutportb(IULA, IULA_BASE) ;
	HLPRoutportb(OPSEL, HLPRinportb(OPSEL) | 0x02) ;
#endif	/* COMMINT */

#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect((INTBASE + INTNUM), (void far *) COMMrcvint) ;

	/* Initialize the V40/V50 interrupt controller */
	HLPRoutportb(IIW1, ICU_IIW1) ;
	HLPRoutportb(IIW2, INTBASE) ;
	HLPRoutportb(IIW4, ICU_IIW4) ;
	HLPRoutportb(IMKW, 0xff) ;
#endif	/* COMMINT */

	/* Initialize timer 1 as the baud rate generator */
	HLPRoutportb(TMD, T1_MODE) ;
	HLPRoutportb(TCT1, lowbyte(BRDIV)) ;
	HLPRoutportb(TCT1, highbyte(BRDIV)) ;

	/* Force the V40/V50 SCU to a known state */
	HLPRoutportb(SCM, (BYTE) 0) ;
	HLPRoutportb(SCM, (BYTE) 0) ;
	HLPRoutportb(SCM, (BYTE) 0) ;
	
	/* Reset the SCU */
	HLPRoutportb(SCM, SCU_RESET) ;

	/* Program the mode register and command registers */
	HLPRoutportb(SMD, SCU_MODE) ;
	HLPRoutportb(SCM, SCU_CMD) ;
	
	/* Enable operation of the remote link */
#if	defined(COMMINT)
	/* Unmask SCU and the ICU interrupt inputs */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) & ~(1 << SIMK_RCV)) ;
	HLPRoutportb(IMKW, HLPRinportb(IMKW) & ~(1 << SCUINT)) ;
#endif	/* COMMINT */
	enable() ;
}
