/*
//		Paradigm Systems Turbo Debugger Interface
//		V40H/V50H Serial Driver (V40H/V50H ICU and polled mode)
//
//		Copyright (C) 1991 Paradigm Systems.  All rights reserved.
//		Portions Copyright (C) 1988 Borland International.  All rights reserved.
//
//		This module provides support for the NEC V40H/V50H SCU (serial control
//		unit) and ICU (interrupt control unit) or in polled mode if the symbol
//		COMMINT is not defined.
//
//		Functions requiring customization are
//			COMMinit				Hardware initialization (SIO/PIC)
//			COMMsendchar		Write a character
//			COMMrcvchar			Read a character
//			COMMrcvint			Receive interrupt handler
//			COMMgotpacket		Tests if a complete packet has been received
*/

#include	<dos.h>							/* Enable/disable functions */

#include "typedefs.h"					/* Common definitions and prototypes */
#include	"target.h"						/* Target system specific information */
#include	"helpers.h"						/* TDREM helper functions */

/*
//		Internal peripheral addresses which must be defined for the V40H/V50H
//		include file.
*/
#define	OPHA_BASE	0x00				/* OPHA initial value */
#define	IULA_BASE	0x20				/* Offset of V40H/V50H ICU */
#define	SULA_BASE	0x60				/* Offset of V40H/V50H SCU */
#define	SC				1					/* I/O port address scale (V40H=1, V50H=2) */
#include	"v40hv50h.h"					/* NEC V40H/V50H definitions */

/*
//		Define the default peripheral register values.
*/
#define	OPCN_INIT	0x06				/* SCU enabled using INT1 */
#define	OPSEL_INIT	0x08				/* SCU enabled */

#define	BAUD9600		104				/* 9600 baud @ 16 Mhz */
#define	BAUD38K		26					/* 38.4K baud @ 16 Mhz */
#define	BAUD115K		9					/* 115.2K baud @ 16 Mhz */

#define	BRC_INIT		BAUD38K			/* Default baud rate */
#define	SCU_MODE		0x4e				/* 8 data, 1 stop, no parity, 16x clock */
#define	SCU_CMD		0x37				/* Enable xmit/rcv */


#if	defined(COMMINT)
#define	IIW1_INIT	0x1d				/* Level sensitive, single mode, IIW4 */
#define	IIW4_INIT	0x01				/* Normal mode, FI command mode */

#define	INTBASE		0x08				/* Hardware interrupt for bus interrupt 0 */
#define	INTNUM		1					/* V40H/V50H serial interrupt */

#if	(COMMINT != INTBASE + INTNUM)
#error	Uninitialized COMM interrupt vector defined in TARGET.H
#endif	/* COMMINT check */
#endif	/* COMMINT */


/*
//		Function:  COMMrcvchar
//
//		Wait for character and read it in.
*/

BYTE		COMMrcvchar(void)
{
	/* Wait for a character to appear */
	while (!(HLPRinportb(SST) & RBRDY))
		 ;

	/* Return it to the caller */
	return	HLPRinportb(SRB) ;
}


/*
//		Function:  COMMsendchar
//
//		Send a character as soon as the transmitter is ready to accept.
*/

void		COMMsendchar(
				BYTE		bTXchar			/* Output character */
			)
{
	/* Poll the port waiting for the transmit buffer to empty */
	while (!(HLPRinportb(SST) & TBRDY))
		;

	/* Write the character to the transmit buffer */
	HLPRoutportb(STB, bTXchar) ;
}


/*
//		Function:  COMMrcvint
//
//		Serial receive interrupt handler.  Starts by getting the first
//		character in the block and then switching off interrupts while the
//		remainder of the block is received.
//
//		WARNING:	No local (class auto) data can be allocated in the interrupt
//					handler since the stack is cleaned up elsewhere if a Ctrl-Break
//					interrupt is processed.  Note that both p and wCount are
//					declared as static and do not occupy space on the stack.
*/

void	interrupt far	COMMrcvint(void)
{
	static BYTE		*p ;					/* Pointer to the packet */
	static WORD		wCount ;				/* Loop counter */


#if	defined(COMMINT)
	/* Mask serial receive interrupts */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) | 0x01) ;
#else
	/* Disable interrupts in polled mode */
	disable() ;
#endif	/* COMMINT */

	/* Initialize the pointer into the packet */
	p = pCOMMpacket ;

	/* Read the character that caused the interrupt */
	*p++ = wCount = HLPRinportb(SRB) ;

	/* Send a runt packet */
	if (wCount > 0)
		COMMsendchar(0) ;

	/* Only process valid packet sizes */
	if (wCount <= COMMBUFLEN + 1)   {
		/* Read the remainder of the block */
		while (wCount--)
			*p++ = COMMrcvchar() ;

		/* Set the packet received flag */
		bGotPacket = 1 ;
	}

#if	defined(COMMINT)
	/* Issue the EOI to interrupt controller */
	HLPRoutportb(IPFW, EOI) ;

	/* Unmask serial receive interrupts */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) & 0xfe) ;
#endif	/* COMMINT */

	/* Execute the packet function - will not return if a Ctrl-Brk */
	Progtestpacket() ;
}


/*
//		Function:  COMMgotpacket
//
//		See if a received packet is waiting.  Returns 1 if a packet is
//		available, returns 0 for no packet.  Note that this function returns
//		1 once, the next call will return 0.
//
//		Note that this routine defines a critical region which must not be
//		violated.
*/

WORD		COMMgotpacket(void)
{
	BYTE		bStatus ;


#if	!defined(COMMINT)
	/* Check if a character is available */
	if (HLPRinportb(SST) & RBRDY)
		COMMrcvint() ;
#endif	/* !COMMINT */

	/* Get the current status, then immediately clear the flag if set */
	if ((bStatus = bGotPacket) == 1)
		bGotPacket = 0 ;

	return	(WORD) bStatus ;
}


/*
//		Function:  COMMinit
//
//		Initialize the communications port by setting the baud rate, installing
//		the receive interrupt handler and performing any other actions.
*/

void		COMMinit(void)
{
	/* No interrupts while initializing the interrupt controller */
	disable() ;

	/* Initialize the V40H/V50H system I/O registers */
	HLPRoutportb(OPSEL, OPSEL_INIT) ;
	HLPRoutportb(OPCN, OPCN_INIT) ;

	/* Set the I/O port addresses of internal peripherals */
	HLPRoutportb(OPHA, OPHA_BASE) ;
	HLPRoutportb(SULA, SULA_BASE) ;
#if	defined(COMMINT)
	HLPRoutportb(IULA, IULA_BASE) ;
	HLPRoutportb(OPSEL, HLPRinportb(OPSEL) | 0x02) ;
#endif	/* COMMINT */

	/* Initialize the SCU baud rate, mode and command registers */
	HLPRoutportb(BRC, BRC_INIT) ;
	HLPRoutportb(SMD, SCU_MODE) ;
	HLPRoutportb(SCM, SCU_CMD) ;

	/* Enable operation of the remote link */
#if	defined(COMMINT)
	/* Load our interrupt routine address */
	HLPRsetvect((INTBASE + INTNUM), (void far *) COMMrcvint) ;

	/* Initialize the V40H/V50H interrupt controller */
	HLPRoutportb(IIW1, IIW1_INIT) ;
	HLPRoutportb(IIW2, INTBASE) ;
	HLPRoutportb(IIW4, IIW4_INIT) ;
	HLPRoutportb(IMKW, 0xff) ;

	/* Unmask SCU and the ICU interrupt inputs */
	HLPRoutportb(SIMK, HLPRinportb(SIMK) & ~(1 << SIMK_RCV)) ;
	HLPRoutportb(IMKW, HLPRinportb(IMKW) & ~(1 << SCUINT)) ;
#endif	/* COMMINT */
	enable() ;
}
